#include <Wire.h>
#include <Keypad.h>

#include <Adafruit_SSD1306.h>

// screen
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 32 // OLED display height, in pixels

#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C

#define MAX_NR_ROWS 2
#define ROW_HEIGHT (SCREEN_HEIGHT / MAX_NR_ROWS)

Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);


// utils
#define abs_f(x) ((x) > 0 ? (x) : -(x))

#define UINT_MAX 4294967295

// motor
#define TIME_BETWEEN_STEPS 4
#define STEPS_PER_REVOLUTION 2048

#define STEPPER_OUT_1 PC3
#define STEPPER_OUT_2 PC2
#define STEPPER_OUT_3 PC1
#define STEPPER_OUT_4 PC0

#define STEPPER_PORT_1 PORTC3
#define STEPPER_PORT_2 PORTC2
#define STEPPER_PORT_3 PORTC1
#define STEPPER_PORT_4 PORTC0

#define STEPPER_REG_W PORTC
#define STEPPER_REG_DIR DDRC

uint8_t stepper_pin_1;
uint8_t stepper_pin_2;
uint8_t stepper_pin_3;
uint8_t stepper_pin_4;

// door
enum DoorState {
  DOOR_OPEN,
  DOOR_CLOSED,
  DOOR_OPENING,
  DOOR_CLOSING
};

DoorState state;

bool should_print_state;

#define PASSWORD_LEN 4

char password[PASSWORD_LEN] = {'1', '2', '3', '4'};

#define DOOR_ANGLE 120.0f

// sound sensor
#define MIC_IN PD3

#define MIC_REG_DIR DDRD

// movement sensor
#define PIR_IN PD2

#define PIR_REG_DIR DDRD

// keypad
const byte ROWS = 4;
const byte COLS = 4;

char keys[ROWS][COLS] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};

byte rowPins[ROWS] = {11, 10, 9, 8};
byte colPins[COLS] = {7, 6, 5, 4};

Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );

void setup() {
  Serial.begin(9600);
  
  // set mic as input
  MIC_REG_DIR &= ~(1 << MIC_IN);

  Serial.println("SETUP");

  // set movement sensor as input
  PIR_REG_DIR &= ~(1 << PIR_IN);

  // initialize display
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));

    // loop forever 
    while(true);
  }

  // initialize text format
  display.setTextSize(2); // Draw 2X-scale text
  display.setTextColor(SSD1306_WHITE);

  // initialize stepper pins as output
  STEPPER_REG_DIR |= (1 << STEPPER_OUT_1);
  STEPPER_REG_DIR |= (1 << STEPPER_OUT_2);
  STEPPER_REG_DIR |= (1 << STEPPER_OUT_3);
  STEPPER_REG_DIR |= (1 << STEPPER_OUT_4);

  stepper_pin_1 = true;
  stepper_pin_2 = stepper_pin_3 = stepper_pin_4 = false;
  
  init_ext_int();

  // initial door state
  state = DOOR_CLOSED;

  should_print_state = true;
}

void init_ext_int() {
  cli();
  
  // set INT1 on falling edge
  EICRA |= (1 << ISC11); // for sound sensor

  // set INT0 on rising edge
  EICRA |= (1 << ISC00) | (1 << ISC01); // for movement sensor

  sei();
}

void start_timer1() {
  cli();

  TIFR1 |= (1 << OCF1A);
  
  TCCR1A = 0;
  TCCR1B = 0;
      
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  
  // 1024 prescaler
  TCCR1B |= (1 << CS12) | (0 << CS11) | (1 << CS10);

  OCR1A = 31249; // = 16000000 / (1024 * 0.5) - 1 (must be < 65536)

  TCNT1 = 0;
  
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);
  
  sei();
}

void toggle_movement() {
  EIFR |= (1 << INTF0); // disable
  EIMSK ^= (1 << INT0);

  start_timer1();

  if (EIMSK & (1 << INT0)) {
    displayCenterTwoRows("MOVEMENT", "ENABLED");

    return;
  }

  displayCenterTwoRows("MOVEMENT", "DISABLED");
}

void toggle_audio() {
  EIFR |= (1 << INTF1); // disable
  EIMSK ^= (1 << INT1);

  start_timer1();

  if (EIMSK & (1 << INT1)) {
    displayCenterTwoRows("AUDIO", "ENABLED");

    return;
  }

  displayCenterTwoRows("AUDIO", "DISABLED");
}

ISR(TIMER1_COMPA_vect) {
  TIMSK1 &= ~(1 << OCIE1A);
  should_print_state = true;
}

void toggle_door() {
  // close door
  if (state == DOOR_OPEN) {
    state = DOOR_CLOSING;

    return;
  }
  
  // open door
  if (state == DOOR_CLOSED) {
    state = DOOR_OPENING;
  }
}

ISR(INT1_vect) {
  toggle_door();
}

ISR(INT0_vect) { 
  toggle_door();
}

void change_password() {
  char c;

  char new_password[PASSWORD_LEN];

  int poz = 0;

  displayCenterTwoRows("ENTER NEW", "PASSWORD");

  while (poz < PASSWORD_LEN) {
    c = keypad.getKey();

    if (!c) {
      continue;
    }
    
    if (!is_digit(c)) {
      start_timer1();
      displayCenterOneRow("ERROR");

      return;
    }

    new_password[poz++] = c;
  }

  for (int i = 0; i < PASSWORD_LEN; ++i) {
    password[i] = new_password[i];
  }

  start_timer1();
  displayCenterTwoRows("PASSWORD", "CHANGED");
}

void enter_password() {
  int poz = 0;
  
  bool correct = true;

  char c;

  displayCenterTwoRows("ENTER", "PASSWORD");

  while (poz < PASSWORD_LEN) {
    c = keypad.getKey();

    if (!c) {
      continue;
    }

    correct &= (c == password[poz++]);
  }

  if (correct) {
    toggle_door();

    return;
  }

  start_timer1();
  displayCenterTwoRows("WRONG", "PASSWORD");
}

bool is_digit(char c) {
  return (c >= '0') && (c <= '9');
}

void loop() {
  char c;

  if (should_print_state) {
    should_print_state = false;

    print_status();
  }

  char key = keypad.getKey();
  
  if (key){

    switch (key) {
      case 'A':
        toggle_movement();
        break;

      case 'B':
        toggle_audio();
        break;

      case 'C':
        change_password();
        break;

      case 'D':
        enter_password();
        break;

      default:
          start_timer1();
          displayCenterTwoRows("WRONG", "COMMAND");
    }

  }

  if (Serial.available() > 0) {
    c = Serial.read();
  }
  
  if (state == DOOR_OPENING) {
    print_status();
    
    rotate_door(-DOOR_ANGLE);

    state = DOOR_OPEN;

    print_status();
  }
    
  if (state == DOOR_CLOSING) {
    print_status();

    rotate_door(DOOR_ANGLE);

    state = DOOR_CLOSED;

    print_status();
  }

  if (c == 'z') {
    rotate_door(-10.0f);
  }

  if (c == 'x') {
    rotate_door(10.0f);
  }

  if (c == 'o') {
    rotate_door(-120.0f);
  }

  if (c == 'c') {
    rotate_door(120.0f);
  }
}

void print_status() {
  if (state == DOOR_OPEN) {
    displayCenterOneRow("OPEN");

    return;
  }

  if (state == DOOR_CLOSED) {
    displayCenterOneRow("CLOSED");

    return;
  }

  if (state == DOOR_OPENING) {
    displayCenterTwoRows("OPENING", "DOOR");

    return;
  }

  if (state == DOOR_CLOSING) {
    displayCenterTwoRows("CLOSING", "DOOR");
  }
}

void displayCenterOneRow(String row) {
  int16_t x;
  int16_t y;
  uint16_t width;
  uint16_t height;

  display.clearDisplay();

  display.getTextBounds(row, 0, 0, &x, &y, &width, &height);

  // display on horizontal and vertical center
  display.setCursor((SCREEN_WIDTH - width) / 2, (SCREEN_HEIGHT - height) / 2);
  display.println(row); // text to display
  
  display.display();
}

void displayCenterTwoRows(String row1, String row2) {
  int16_t x;
  int16_t y;
  uint16_t width;
  uint16_t height;

  display.clearDisplay();
  
  display.getTextBounds(row1, 0, 0, &x, &y, &width, &height);

  // display on horizontal and vertical center
  display.setCursor((SCREEN_WIDTH - width) / 2, 0);
  display.println(row1); // row1 to display

  display.getTextBounds(row2, 0, 0, &x, &y, &width, &height);
  
  display.setCursor((SCREEN_WIDTH - width) / 2, SCREEN_HEIGHT / 2);
  display.println(row2); // row1 to display

  display.display();
}

void rotate_door(float degree) {  
  int steps = (abs_f(degree) / 360.0f) * STEPS_PER_REVOLUTION;

  Serial.println(steps);
  
  void (*f)() = makeStepCW;

  if (degree < 0) {
    f = makeStepCCW;
  }

  for (int i = 0; i < steps; ++i) {
    f();
    delay(TIME_BETWEEN_STEPS);
  }
}

void makeStepCW() {
  uint8_t temp;

  temp = stepper_pin_1;
  stepper_pin_1 = stepper_pin_2;
  stepper_pin_2 = stepper_pin_3;
  stepper_pin_3 = stepper_pin_4;
  stepper_pin_4 = temp;

  write_motor_pins();
}

void makeStepCCW() {
  uint8_t temp;

  temp = stepper_pin_4;
  stepper_pin_4 = stepper_pin_3;
  stepper_pin_3 = stepper_pin_2;
  stepper_pin_2 = stepper_pin_1;
  stepper_pin_1 = temp;

  write_motor_pins();
}

void write_motor_pins() {
  write_motor_pin(STEPPER_PORT_1, stepper_pin_1);
  write_motor_pin(STEPPER_PORT_2, stepper_pin_2);
  write_motor_pin(STEPPER_PORT_3, stepper_pin_3);
  write_motor_pin(STEPPER_PORT_4, stepper_pin_4);
}

void write_motor_pin(uint8_t pin, uint8_t value) {
  if (value == 0) {
    STEPPER_REG_W &= ~(1 << pin);

    return;
  }

  if (value == 1) {
    STEPPER_REG_W |= (1 << pin);
  }
}
